@@ -1,4 +1,7 @@
 import math
+from random import random
+from functools import wraps, partial
+
 import torch
 import torch.nn.functional as F
 from torch import nn, einsum
@@ -90,17 +93,17 @@ class Attention(nn.Module):
         self,
         dim_head,
         causal = True,
-        cosine_sim = False,
         qk_rmsnorm = False,
-        cosine_sim_scale = 8
+        qk_rmsnorm_scale = 8,
+        single_head_kv = True
     ):
         super().__init__()
         self.causal = causal
 
-        assert not (cosine_sim and qk_rmsnorm)
-        self.cosine_sim = cosine_sim
         self.qk_rmsnorm = qk_rmsnorm
-        self.cosine_sim_scale = cosine_sim_scale
+        self.qk_rmsnorm_scale = qk_rmsnorm_scale
+
+        self.single_head_kv = single_head_kv
 
         if qk_rmsnorm:
             self.q_scale = nn.Parameter(torch.ones(dim_head))
@@ -110,15 +113,15 @@ class Attention(nn.Module):
 
         scale = q.shape[-1] ** -0.5
 
-        if self.cosine_sim or self.qk_rmsnorm:
+        if self.qk_rmsnorm:
             q, k = map(l2norm, (q, k))
-            scale = self.cosine_sim_scale
+            scale = self.qk_rmsnorm_scale
 
         if self.qk_rmsnorm:
             q = q * self.q_scale
             k = k * self.k_scale
 
-        kv_einsum = '... j d' if k.ndim == 3 else '... h j d'
+        kv_einsum = '... j d' if self.single_head_kv else '... h j d'
 
         sim = einsum(f'... h i d, {kv_einsum} -> ... h i j', q, k) * scale
 
@@ -149,10 +152,10 @@ class AttentionBlock(nn.Module):
         causal = True,
         dim_head = 64,
         heads = 8,
-        cosine_sim = False,
         qk_rmsnorm = False,
-        cosine_sim_scale = 8,
+        qk_rmsnorm_scale = 8,
         num_state_vectors = 0,
+        single_head_kv = True
     ):
         super().__init__()
         inner_dim = dim_head * heads
@@ -160,8 +163,12 @@ class AttentionBlock(nn.Module):
 
         self.norm = LayerNorm(dim)
 
-        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias = False)
-        self.attn = Attention(dim_head, causal = causal, cosine_sim = cosine_sim, qk_rmsnorm = qk_rmsnorm, cosine_sim_scale = cosine_sim_scale)
+        self.to_q = nn.Linear(dim, inner_dim, bias = False)
+
+        self.single_head_kv = single_head_kv
+        self.to_kv = nn.Linear(dim, (inner_dim if not single_head_kv else dim_head) * 2, bias = False)
+
+        self.attn = Attention(dim_head, causal = causal, qk_rmsnorm = qk_rmsnorm, qk_rmsnorm_scale = qk_rmsnorm_scale, single_head_kv = single_head_kv)
 
         self.block_width = block_width
         self.is_recurrent_layer = num_state_vectors > 0
@@ -174,16 +181,18 @@ class AttentionBlock(nn.Module):
             self.q_to_state = nn.Linear(dim, inner_dim, bias = False)
             self.q_from_state = nn.Linear(dim, inner_dim, bias = False)
 
-            self.state_to_qkv = nn.Linear(dim, inner_dim * 3, bias = False)
+            self.state_to_q = nn.Linear(dim, inner_dim, bias = False)
+            self.state_to_kv = nn.Linear(dim, (inner_dim if not single_head_kv else dim_head) * 2, bias = False)
+
             self.init_state = nn.Parameter(torch.randn(num_state_vectors, dim))
             self.state_pos_ids = nn.Parameter(torch.randn(num_state_vectors, dim))
 
             self.to_state_out = nn.Linear(inner_dim * 2, dim, bias = False)
 
-            self.to_state_cross_attn = Attention(dim_head, causal = False, cosine_sim = cosine_sim, qk_rmsnorm = qk_rmsnorm, cosine_sim_scale = cosine_sim_scale)
+            self.to_state_cross_attn = Attention(dim_head, causal = False, qk_rmsnorm = qk_rmsnorm, qk_rmsnorm_scale = qk_rmsnorm_scale, single_head_kv = single_head_kv)
 
-            self.state_self_attn = Attention(dim_head, causal = False, cosine_sim = cosine_sim, qk_rmsnorm = qk_rmsnorm, cosine_sim_scale = cosine_sim_scale)
-            self.from_state_cross_attn = Attention(dim_head, causal = False, cosine_sim = cosine_sim, qk_rmsnorm = qk_rmsnorm, cosine_sim_scale = cosine_sim_scale)
+            self.state_self_attn = Attention(dim_head, causal = False, qk_rmsnorm = qk_rmsnorm, qk_rmsnorm_scale = qk_rmsnorm_scale, single_head_kv = single_head_kv)
+            self.from_state_cross_attn = Attention(dim_head, causal = False, qk_rmsnorm = qk_rmsnorm, qk_rmsnorm_scale = qk_rmsnorm_scale, single_head_kv = single_head_kv)
 
             # gating related parameters - using the fixed simple config
 
@@ -214,12 +223,17 @@ class AttentionBlock(nn.Module):
 
         # queries, keys, values and split out heads
 
-        q, k, v = self.to_qkv(x).chunk(3, dim = -1)
-        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = self.heads), (q, k, v))
+        q, k, v = (self.to_q(x), *self.to_kv(x).chunk(2, dim = -1))
+
+        split_head = partial(rearrange, pattern = 'b n (h d) -> b h n d', h = self.heads)
+        q = split_head(q)
+
+        if not self.single_head_kv:
+            k, v = map(split_head, (k, v))
 
         # bucket the queries, keys, values by block width
 
-        bq, bk, bv = map(lambda t: rearrange(t, 'b h (w n) d -> b w h n d', n = width), (q, k, v))
+        bq, bk, bv = map(lambda t: rearrange(t, 'b ... (w n) d -> b w ... n d', n = width), (q, k, v))
 
         # save the last key / values as memories for recurrence
 
@@ -231,7 +245,7 @@ class AttentionBlock(nn.Module):
         if exists(xl_memories):
             # if past memories are passed in, concat as the first bucket
             past_k, past_v = xl_memories
-            past_k, past_v = map(lambda t: rearrange(t, 'b h n d -> b 1 h n d'), (past_k, past_v))
+            past_k, past_v = map(lambda t: rearrange(t, 'b ... n d -> b 1 ... n d'), (past_k, past_v))
             bk = torch.cat((past_k, bk), dim = 1)
             bv = torch.cat((past_v, bv), dim = 1)
         else:
@@ -289,8 +303,11 @@ class AttentionBlock(nn.Module):
 
             # self attention qkv for states
 
-            states = self.state_to_qkv(self.init_state).chunk(3, dim = -1)
-            state_q, state_k, state_v = map(lambda t: repeat(t, 'n (h d) -> b h n d', h = self.heads, b = batch), states)
+            state_q, state_k, state_v = (self.state_to_q(self.init_state), *self.state_to_kv(self.init_state).chunk(2, dim = -1))
+            state_q = repeat(state_q, 'n (h d) -> b h n d', h = self.heads, b = batch)
+
+            if not self.single_head_kv:
+                state_k, state_v = map(lambda t: repeat(t, 'n (h d) -> b h n d', h = self.heads, b = batch), (state_k, state_v))
 
             # cross attend to the past states key values
 
@@ -339,20 +356,20 @@ class BlockRecurrentTransformer(nn.Module):
         depth,
         dim_head = 64,
         heads = 8,
-        cosine_sim = False,
-        qk_rmsnorm = True,
-        cosine_sim_scale = 8,
+        single_head_kv = True,
+        all_layers_qk_rmsnorm = False,
         ff_mult = 4,
-        ignore_index = -100,
         max_seq_len = 1024,
         block_width = 512,
         xl_memories_layers: Optional[Tuple[int, ...]] = None,
         recurrent_layers: Optional[Tuple[int, ...]] = None,
-        num_state_vectors = 512,
+        num_state_vectors = None,
         dynamic_pos_bias_dim = None,
-        enhanced_recurrence = False
+        enhanced_recurrence = False,
+        ignore_index = -100
     ):
         super().__init__()
+        num_state_vectors = default(num_state_vectors, block_width)
         xl_memories_layers = default(xl_memories_layers, tuple(range(1, depth + 1)))
         self.xl_memories_layers = set(xl_memories_layers)
 
@@ -378,8 +395,17 @@ class BlockRecurrentTransformer(nn.Module):
 
         for layer in range(1, depth + 1):
             is_recurrent_layer = layer in self.recurrent_layers
+            is_xl_layer = layer in self.xl_memories_layers
+
             layer_num_state_vectors = num_state_vectors if is_recurrent_layer else 0
 
+            # only layers with xl memories
+            # or has recurrence in horizontal direction
+            # use qk rmsnorm (in paper, they use cosine sim attention, but i think qk rmsnorm is more proven given Vit 22B paper)
+            # one can also override to use all qk rmsnorm by setting all_layers_qk_rmsnorm = True
+
+            qk_rmsnorm = all_layers_qk_rmsnorm or is_recurrent_layer or is_xl_layer
+
             self.layers.append(nn.ModuleList([
                 AttentionBlock(
                     dim,
@@ -387,9 +413,8 @@ class BlockRecurrentTransformer(nn.Module):
                     block_width = block_width,
                     dim_head = dim_head,
                     heads = heads,
-                    cosine_sim = cosine_sim,
+                    single_head_kv = single_head_kv,
                     qk_rmsnorm = qk_rmsnorm,
-                    cosine_sim_scale = cosine_sim_scale,
                     num_state_vectors = layer_num_state_vectors
                 ),
                 FeedForward(dim, mult = ff_mult)
@@ -577,12 +602,17 @@ class BlockRecurrentTransformer(nn.Module):
 class RecurrentTrainerWrapper(nn.Module):
     def __init__(
         self,
-        transformer: BlockRecurrentTransformer
+        transformer: BlockRecurrentTransformer,
+        xl_memories_dropout = 0.,
+        state_dropout = 0.
     ):
         super().__init__()
         self.transformer = transformer
         self.seq_len = transformer.max_seq_len
 
+        self.xl_memories_dropout = xl_memories_dropout
+        self.state_dropout = state_dropout
+
     @eval_decorator
     @torch.no_grad()
     def generate(
@@ -591,6 +621,7 @@ class RecurrentTrainerWrapper(nn.Module):
         length,
         **kwargs
     ):
+        seq_len = self.seq_len
         start_len = prime.shape[-1]
         assert start_len < length
 
@@ -600,15 +631,22 @@ class RecurrentTrainerWrapper(nn.Module):
         memories = []
         states = []
 
-        while output.shape[-1] < length:
+        # determine lengths
 
-            if (length - output.shape[-1]) <= self.seq_len:
-                next_length = length % self.seq_len
-            else:
-                next_length = self.seq_len + 1
+        has_remainder = not divisible_by(length, seq_len)
+        remainder_amount = length % seq_len
+        total_segments = math.ceil(length / seq_len)
 
-            if next_length == 0:
-                break
+        if not has_remainder:
+            lengths = (*((seq_len + 1,) * (total_segments - 1)), seq_len)
+        elif remainder_amount == 1:
+            lengths = (seq_len + 1,) * (total_segments - 1)
+        else:
+            lengths = (*((seq_len + 1,) * (total_segments - 1)), remainder_amount)
+
+        # loop through lengths
+
+        for next_length in lengths:
 
             segment_output, memories, states = self.transformer.generate(
                 output[:, -current_len:],
@@ -639,6 +677,12 @@ class RecurrentTrainerWrapper(nn.Module):
             start = ind * seq_len
             end = start + seq_len + 1
 
+            if self.training and random() < self.xl_memories_dropout:
+                memories.clear()
+
+            if self.training and random() < self.state_dropout:
+                states.clear()
+
             loss, memories, states = self.transformer(
                 x[:, start:end],
                 xl_memories = memories,
